@@ -4,7 +4,7 @@
 
 # Colour the widget for the front to back card type red.
 
-from PyQt4 import QtGui
+from PyQt5 import QtGui
 
 from mnemosyne.libmnemosyne.plugin import Plugin
 from mnemosyne.libmnemosyne.card_types.front_to_back import FrontToBack
@@ -25,10 +25,8 @@ class RedCardTypeWdgt(GenericCardTypeWdgt):
     component_type = "card_type_widget"
     used_for = FrontToBack
 
-    def __init__(self, component_manager, parent):
-        card_type = FrontToBack(component_manager)
-        GenericCardTypeWdgt.__init__(self, component_manager,
-                                     parent, card_type)
+    def __init__(self, **kwds):
+        super().__init__(**kwds)
         for edit_box in self.edit_boxes:
             p = QtGui.QPalette()
             p.setColor(QtGui.QPalette.Active, QtGui.QPalette.Base, \
