@@ -45,3 +45,6 @@ class Fact(CompareOnId):
     def __setitem__(self, key, value):
         self.data[key] = value
         
+    def __contains__(self, key):
+        return key in self.data
+        
