@@ -2,24 +2,24 @@
 # clone_card_type_dlg.py <Peter.Bienstman@UGent.be>
 #
 
-from PyQt4 import QtCore, QtGui
+from PyQt5 import QtCore, QtGui, QtWidgets
 
 from mnemosyne.libmnemosyne.translator import _
 from mnemosyne.libmnemosyne.component import Component
 from mnemosyne.pyqt_ui.ui_clone_card_type_dlg import Ui_CloneCardTypeDlg
 
 
-class CloneCardTypeDlg(QtGui.QDialog, Ui_CloneCardTypeDlg, Component):
+class CloneCardTypeDlg(QtWidgets.QDialog, Component, Ui_CloneCardTypeDlg):
 
-    def __init__(self, parent, component_manager):
-        Component.__init__(self, component_manager)
-        QtGui.QDialog.__init__(self, parent)
+    def __init__(self, **kwds):
+        super().__init__(**kwds)
         self.setupUi(self)
         self.setWindowFlags(self.windowFlags() \
             | QtCore.Qt.WindowMinMaxButtonsHint)
         self.setWindowFlags(self.windowFlags() \
             & ~ QtCore.Qt.WindowContextHelpButtonHint)
-        for card_type in self.card_types():
+        self.sorted_card_types = self.database().sorted_card_types()
+        for card_type in self.sorted_card_types:
             self.parent_type.addItem(_(card_type.name))
 
     def name_changed(self):
@@ -29,10 +29,10 @@ class CloneCardTypeDlg(QtGui.QDialog, Ui_CloneCardTypeDlg, Component):
             self.OK_button.setEnabled(True)
 
     def accept(self):
-        parent_instance = self.card_types()[self.parent_type.currentIndex()]
-        clone_name = unicode(self.name.text())
+        parent_instance = self.sorted_card_types[self.parent_type.currentIndex()]
+        clone_name = self.name.text()
         clone = self.controller().clone_card_type(\
             parent_instance, clone_name)
         if not clone:
             return
-        QtGui.QDialog.accept(self)
+        QtWidgets.QDialog.accept(self)
