@@ -6,39 +6,17 @@
 
 import os
 import socket
-import urllib
+import urllib.request, urllib.parse, urllib.error
 import tarfile
-import httplib
+import http.client
 
-from partner import Partner
-from text_formats.xml_format import XMLFormat
-from utils import traceback_string, SyncError, SeriousSyncError
-
-# Avoid delays caused by Nagle's algorithm.
-# http://www.cmlenz.net/archives/2008/03/python-httplib-performance-problems
-
-realsocket = socket.socket
-def socketwrap(family=socket.AF_INET, type=socket.SOCK_STREAM, proto=0):
-    sockobj = realsocket(family, type, proto)
-    sockobj.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
-    return sockobj
-socket.socket = socketwrap
-
-# Buffer the response socket.
-# http://mail.python.org/pipermail/python-bugs-list/2006-September/035156.html
-# Fix included here for systems running Python 2.5.
-
-class HTTPResponse(httplib.HTTPResponse):
-
-    def __init__(self, sock, **kw):
-        httplib.HTTPResponse.__init__(self, sock, **kw)
-        self.fp = sock.makefile("rb") # Was unbuffered: sock.makefile("rb", 0)
-
-httplib.HTTPConnection.response_class = HTTPResponse
+from .partner import Partner
+from .text_formats.xml_format import XMLFormat
+from .utils import SyncError, SeriousSyncError, traceback_string
 
 # Register binary formats.
 
-from binary_formats.mnemosyne_format import MnemosyneFormat
+from .binary_formats.mnemosyne_format import MnemosyneFormat
 BinaryFormats = [MnemosyneFormat]
 
 
@@ -95,8 +73,8 @@ class Client(Partner):
 
         # If we haven't done so, determine whether we're behind a proxy.
         if self.behind_proxy is None:
-            import urllib
-            proxies = urllib.getproxies()
+            import urllib.request, urllib.parse, urllib.error
+            proxies = urllib.request.getproxies()
             if "http" in proxies:
                 self.behind_proxy = True
                 self.proxy = proxies["http"]
@@ -104,18 +82,18 @@ class Client(Partner):
                 self.behind_proxy = False
         # Create a new connection or reuse an existing one.
         if self.behind_proxy:
-            httplib.HTTPConnection._http_vsn = 10
-            httplib.HTTPConnection._http_vsn_str = "HTTP/1.0"
+            http.client.HTTPConnection._http_vsn = 10
+            http.client.HTTPConnection._http_vsn_str = "HTTP/1.0"
             if self.proxy is not None:
-                self.con = httplib.HTTPConnection(self.proxy, self.port)
+                self.con = http.client.HTTPConnection(self.proxy, self.port)
             else:  # Testsuite has set self.behind_proxy to True to simulate
                 # being behind a proxy.
-                self.con = httplib.HTTPConnection(self.server, self.port)
+                self.con = http.client.HTTPConnection(self.server, self.port)
         else:
-            httplib.HTTPConnection._http_vsn = 11
-            httplib.HTTPConnection._http_vsn_str = "HTTP/1.1"
+            http.client.HTTPConnection._http_vsn = 11
+            http.client.HTTPConnection._http_vsn_str = "HTTP/1.1"
             if not self.con:
-                self.con = httplib.HTTPConnection(self.server, self.port)
+                self.con = http.client.HTTPConnection(self.server, self.port)
 
     def url(self, url_string):
         if self.behind_proxy and self.proxy:
@@ -135,11 +113,18 @@ class Client(Partner):
             if self.check_for_edited_local_media_files:
                 self.ui.set_progress_text("Checking for edited media files...")
                 self.database.check_for_edited_media_files()
-            socket.setdefaulttimeout(20)
+            # Always create media files, otherwise they are not synced across.
+            self.ui.set_progress_text("Dynamically creating media files...")
+            self.database.dynamically_create_media_files()
+            # Set timeout long enough for e.g. a slow NAS waking from 
+            # hibernation.
+            socket.setdefaulttimeout(60)
             self.login(username, password)
-            # Generating media files at the server side could take some time.
+            # Generating media files at the server side could take some time,
+            # so we update the timeout.
             self.con = None
             socket.setdefaulttimeout(15*60)
+            self.get_server_check_media_files()
             # Do a full sync after either the client or the server has restored
             # from a backup.
             if self.database.is_sync_reset_needed(\
@@ -148,19 +133,32 @@ class Client(Partner):
                 self.resolve_conflicts(restored_from_backup=True)
             # First sync, fetch database from server.
             elif self.database.is_empty():
-                self.get_server_media_files()
+                self.get_server_media_files(redownload_all=True)
+                self.get_server_archive_files()
                 if self.server_info["supports_binary_transfer"]:
                     self.get_server_entire_database_binary()
                 else:
                     self.get_server_entire_database()
                 self.get_sync_finish()
+                # Fetch config settings.
+                self.login(username, password)
+                self.get_server_generate_log_entries_for_settings()
+                self.get_server_log_entries()
+                self.get_sync_finish()
             # First sync, put binary database to server if supported.
             elif not self.database.is_empty() and \
                     self.server_info["is_database_empty"] and \
                     self.supports_binary_upload():
                 self.put_client_media_files(reupload_all=True)
+                self.put_client_archive_files()
                 self.put_client_entire_database_binary()
                 self.get_sync_finish()
+                # Upload config settings.
+                self.login(username, password)
+                self.database.generate_log_entries_for_settings()
+                self.put_client_log_entries()
+                self.get_server_log_entries()
+                self.get_sync_finish()
             else:
                 # Upload local changes and check for conflicts.
                 result = self.put_client_log_entries()
@@ -172,7 +170,7 @@ class Client(Partner):
                 else:
                     self.resolve_conflicts()
             self.ui.show_information("Sync finished!")
-        except Exception, exception:
+        except Exception as exception:
             self.ui.close_progress()
             serious = True
             if type(exception) == type(socket.gaierror()):
@@ -195,26 +193,33 @@ class Client(Partner):
                 # sync next time.
                 self.ui.show_error("Sync failed, restoring from backup. " + \
                     "The next sync will need to be a full sync.")
-                self.database.restore(backup_file)
+                if backup_file:
+                    self.database.restore(backup_file)
         finally:
+            
+            # TMP: preparation for future fix
+            #self.get_sync_cancel()
+            
+            
             if self.con:
                 self.con.close()
             self.ui.close_progress()
 
-    def supports_binary_upload(self):
+    def supports_binary_upload(self):        
         return self.capabilities == "mnemosyne_dynamic_cards" and \
             self.interested_in_old_reps and self.store_pregenerated_data \
             and self.program_name == self.server_info["program_name"] and \
             self.program_version == self.server_info["program_version"]
 
     def resolve_conflicts(self, restored_from_backup=False):
-        if restored_from_backup:
+        if restored_from_backup:    
             message = "The database was restored from a backup, either " + \
-                "automatically because of an aborted sync or manually by " + \
-                "the user.\nFor safety, a full sync needs to happen and " + \
-                "you need to choose which copy of the database to keep.\n"
+    "automatically because of an aborted sync or manually by " + \
+    "the user.\nFor safety, a full sync of cards and history needs to happen and " + \
+    "you need to choose which copy of the database to keep and " + \
+    "which copy to discard.\n"
         else:
-            message = "Conflicts detected during sync!"
+            message = "Conflicts detected during sync! This typically happens if you review the same card on both machines. Choose which version of the database (cards + history) to keep and which version to discard."
         # Ask for conflict resolution direction.
         if self.supports_binary_upload():
             result = self.ui.show_question(message,
@@ -233,6 +238,7 @@ class Client(Partner):
         # with a third party will also trigger a full sync.
         if result == "KEEP_REMOTE":
             self.get_server_media_files(redownload_all=True)
+            self.get_server_archive_files()
             if self.server_info["supports_binary_transfer"]:
                 self.get_server_entire_database_binary()
             else:
@@ -242,6 +248,7 @@ class Client(Partner):
         # Keep local.
         elif result == "KEEP_LOCAL":
             self.put_client_media_files(reupload_all=True)
+            self.put_client_archive_files()
             self.put_client_entire_database_binary()
             self.get_sync_finish()
         # Cancel.
@@ -250,8 +257,9 @@ class Client(Partner):
 
     def _check_response_for_errors(self, response, can_consume_response=True):
         # Check for non-Mnemosyne error messages.
-        if response.status != httplib.OK:
-            raise SeriousSyncError("Internal server error:\n" + response.read())
+        if response.status != http.client.OK:
+            raise SeriousSyncError("Internal server error:\n" + \
+                                   str(response.read(), "utf-8"))
         if can_consume_response == False:
             return
         # Check for Mnemosyne error messages.
@@ -288,9 +296,10 @@ class Client(Partner):
             self.request_connection()
             self.con.request("PUT", self.url("/login"),
                 self.text_format.repr_partner_info(client_info).\
-                encode("utf-8") + "\n")
+                encode("utf-8") + b"\n")
             response = self.con.getresponse()
-        except Exception, e:
+        except Exception as e:
+            print(e)
             raise SyncError("Could not connect to server!")
         # Check for errors, but don't force a restore from backup if we can't
         # login.
@@ -299,7 +308,7 @@ class Client(Partner):
                 response, can_consume_response=False)
         except SeriousSyncError:
             raise SyncError("Logging in: server error.")
-        response = response.read()
+        response = str(response.read(), "utf-8")
         if "message" in response:
             message, traceback = self.text_format.parse_message(response)
             message = message.lower()
@@ -320,11 +329,27 @@ class Client(Partner):
         elif self.server_info["user_id"] != client_info["user_id"] and \
             self.server_info["is_database_empty"] == False:
             raise SyncError("Error: mismatched user ids.\n" + \
-                "The first sync should happen on an empty database.")
+                "The first sync should happen on an empty database.\n" + \
+                "Backup then delete the local database and try again.")
+        if self.server_info["database_version"] != client_info["database_version"]:
+            raise SyncError("Error: database version mismatch.\n" + \
+                "Make sure you are running the latest Mnemosyne version on both devices involved in the sync.")
+        if self.server_info["program_version"] != client_info["program_version"]:
+            raise SyncError("Error: Mnemosyne version mismatch.\n" + \
+                "Make sure you are running the latest Mnemosyne version on both devices involved in the sync.")         
         self.database.create_if_needed_partnership_with(\
             self.server_info["machine_id"])
         self.database.merge_partners(self.server_info["partners"])
 
+    def get_server_check_media_files(self):
+        self.ui.set_progress_text("Asking server to check for updated media files...")
+        self.request_connection()
+        self.con.request("GET", self.url(\
+            "/server_check_media_files?" + \
+            "session_token=%s" % (self.server_info["session_token"], )))
+        response = self.con.getresponse()
+        self._check_response_for_errors(response, can_consume_response=True)
+
     def put_client_log_entries(self):
 
         """Contrary to binary files, the size of the log is not known until we
@@ -399,7 +424,7 @@ class Client(Partner):
 
     def _download_log_entries(self, stream):
         element_loop = self.text_format.parse_log_entries(stream)
-        number_of_entries = int(element_loop.next())
+        number_of_entries = int(next(element_loop))
         if number_of_entries == 0:
             return
         self.ui.set_progress_range(number_of_entries)
@@ -462,26 +487,64 @@ class Client(Partner):
             self.server_info["machine_id"])
         self.database.remove_partnership_with(self.machine_id)
 
+    def get_server_generate_log_entries_for_settings(self):
+        self.ui.set_progress_text("Getting settings...")
+        self.request_connection()
+        self.con.request("GET", self.url(\
+            "/server_generate_log_entries_for_settings?" + \
+            "session_token=%s" % (self.server_info["session_token"], )))
+        response = self.con.getresponse()
+        self._check_response_for_errors(response, can_consume_response=True)
+
     def put_client_media_files(self, reupload_all=False):
         self.ui.set_progress_text("Sending media files...")
+        # Get list of filenames in the format <mediadir>/<filename>, i.e.
+        # relative to the data_dir. Note we always use / internally. 
+        subdir = os.path.basename(self.database.media_dir())
         if reupload_all:
-            filenames = list(self.database.all_media_filenames())
+            filenames = [subdir + "/" + filename for filename in \
+                         self.database.all_media_filenames()]
         else:
-            filenames = list(self.database.media_filenames_to_sync_for(\
-                self.server_info["machine_id"]))
+            filenames = [subdir + "/" + filename for filename in \
+                         self.database.media_filenames_to_sync_for(\
+                         self.server_info["machine_id"])]
+        # Calculate file size and upload.
         total_size = 0
         for filename in filenames:
             total_size += os.path.getsize(os.path.join(\
-                self.database.media_dir(), filename))
+                self.database.data_dir(), filename))
+        self.put_client_binary_files(filenames, total_size)
+        self.ui.close_progress()
+          
+    def put_client_archive_files(self):  
+        archive_dir = os.path.join(self.database.data_dir(), "archive")
+        if not os.path.exists(archive_dir):
+            return
+        # Get list of filenames in the format "archive"/<filename>, i.e.
+        # relative to the data_dir. Note we always use / internally. 
+        self.ui.set_progress_text("Sending archive files...")       
+        filenames = ["archive/" + filename for filename in \
+                     os.listdir(archive_dir) if os.path.isfile\
+                     (os.path.join(archive_dir, filename))]
+        # Calculate file size and upload.
+        total_size = 0
+        for filename in filenames:
+            total_size += os.path.getsize(os.path.join(\
+                self.database.data_dir(), filename))
+        self.put_client_binary_files(filenames, total_size)
+        self.ui.close_progress()        
+            
+    def put_client_binary_files(self, filenames, total_size):   
+        # Filenames are relative to the data_dir.
         self.ui.set_progress_range(total_size)
         self.ui.set_progress_update_interval(total_size/50)
         for filename in filenames:
             self.request_connection()
             self.con.putrequest("PUT",
-                self.url("/client_media_file?session_token=%s&filename=%s" \
+                self.url("/client_binary_file?session_token=%s&filename=%s" \
                 % (self.server_info["session_token"],
-                urllib.quote(filename.encode("utf-8"), ""))))
-            full_path = os.path.join(self.database.media_dir(), filename)
+                urllib.parse.quote(filename.encode("utf-8"), ""))))
+            full_path = os.path.join(self.database.data_dir(), filename)
             file_size = os.path.getsize(full_path)
             self.con.putheader("content-length", file_size)
             self.con.endheaders()
@@ -492,7 +555,9 @@ class Client(Partner):
         self.ui.set_progress_value(total_size)
 
     def get_server_media_files(self, redownload_all=False):
+        self.ui.set_progress_text("Getting list of media files to download...")
         # Get list of names of all media files to download.
+        # Filenames are relative to the data_dir.
         media_url = "/server_media_filenames?session_token=%s" \
             % (self.server_info["session_token"], )
         if redownload_all:
@@ -507,25 +572,54 @@ class Client(Partner):
             # since we reuse our connection.
             response.read()
             return
-        # Download each media file.
+        filenames = []
+        for filename in response.read().split(b"\n"):
+            filenames.append(str(filename, "utf-8"))
         self.ui.set_progress_text("Getting media files...")
+        self.get_server_binary_files(filenames, total_size)    
+        self.ui.close_progress()
+  
+    def get_server_archive_files(self):
+        self.ui.set_progress_text("Getting list of archive files to download...")
+        # Get list of names of all archive files to download.
+        # Filenames are relative to the data_dir.
+        archive_url = "/server_archive_filenames?session_token=%s" \
+            % (self.server_info["session_token"], )
+        self.request_connection()
+        self.con.request("GET", self.url(archive_url))
+        response = self.con.getresponse()
+        self._check_response_for_errors(response, can_consume_response=False)
+        total_size = int(response.getheader("mnemosyne-content-length"))
+        if total_size == 0:
+            # Make sure to read the full message, even if it's empty,
+            # since we reuse our connection.
+            response.read()
+            return
+        filenames = []
+        for filename in response.read().split(b"\n"):
+            filenames.append(str(filename, "utf-8"))
+        self.ui.set_progress_text("Getting archive files...")
+        self.get_server_binary_files(filenames, total_size)    
+        self.ui.close_progress()  
+               
+    def get_server_binary_files(self, filenames, total_size):
         self.ui.set_progress_range(total_size)
-        self.ui.set_progress_update_interval(total_size/50)
-        for filename in response.read().split("\n"):
-            filename = unicode(filename, "utf-8")
+        self.ui.set_progress_update_interval(total_size/50)  
+        for filename in filenames:
             self.request_connection()
             self.con.request("GET",
-                self.url("/server_media_file?session_token=%s&filename=%s" \
+                self.url("/server_binary_file?session_token=%s&filename=%s" \
                 % (self.server_info["session_token"],
-                urllib.quote(filename.encode("utf-8"), ""))))
+                urllib.parse.quote(filename.encode("utf-8"), ""))))
             response = self.con.getresponse()
             self._check_response_for_errors(response,
                 can_consume_response=False)
             file_size = int(response.getheader("mnemosyne-content-length"))
             # Make sure a malicious server cannot overwrite anything outside
             # of the media directory.
-            filename = filename.replace("..", "")
-            filename = os.path.join(self.database.media_dir(), filename)
+            filename = filename.replace("../", "").replace("..\\", "")
+            filename = filename.replace("/..", "").replace("\\..", "")
+            filename = os.path.join(self.database.data_dir(), filename)
             self.download_binary_file(response, filename,
                                       file_size, progress_bar=False)
             self.ui.increase_progress(file_size)
@@ -534,9 +628,9 @@ class Client(Partner):
     def get_sync_cancel(self):
         self.ui.set_progress_text("Cancelling sync...")
         self.request_connection()
+        session_token = self.server_info.get("session_token", "none")
         self.con.request("GET", self.url("/sync_cancel?session_token=%s" \
-            % (self.server_info["session_token"], )),
-            headers={"connection": "close"})
+            % (session_token, )), headers={"connection": "close"})          
         self._check_response_for_errors(self.con.getresponse())
 
     def get_sync_finish(self):
